@@ -10,10 +10,6 @@
 // For compilers that support precompilation, includes "wx/wx.h".
 #include "wx/wxprec.h"
 
-#ifdef __BORLANDC__
-#pragma hdrstop
-#endif
-
 #if wxUSE_GTKPRINT
 
 #include "wx/gtk/print.h"
@@ -34,7 +30,7 @@
 #include "wx/paper.h"
 #include "wx/modalhook.h"
 
-#include <gtk/gtk.h>
+#include "wx/gtk/private/wrapgtk.h"
 
 #if GTK_CHECK_VERSION(2,14,0)
 #include <gtk/gtkunixprint.h>
@@ -185,9 +181,7 @@ static GtkPaperSize* wxGetGtkPaperSize(wxPaperSize paperId, const wxSize& size)
         return gtk_paper_size_new(gtk_paper_size_get_default());
 
 #if GTK_CHECK_VERSION(2,12,0)
-#ifndef __WXGTK3__
-    if (gtk_check_version(2,12,0) == NULL)
-#endif
+    if (wx_is_at_least_gtk2(12))
     {
         // look for a size match in GTK's GtkPaperSize list
         const double w = size.x;
@@ -216,7 +210,7 @@ static GtkPaperSize* wxGetGtkPaperSize(wxPaperSize paperId, const wxSize& size)
     // last resort, use a custom GtkPaperSize
     const wxString title = _("Custom size");
     char name[40];
-    g_snprintf(name, sizeof(name), "custom_%ux%u", size.x, size.y);
+    g_snprintf(name, sizeof(name), "custom_%dx%d", size.x, size.y);
     return gtk_paper_size_new_custom(
         name, title.utf8_str(), size.x, size.y, GTK_UNIT_MM);
 }
@@ -233,25 +227,23 @@ public:
     wxGtkPrintModule()
     {
     }
-    bool OnInit();
-    void OnExit() {}
+    bool OnInit() wxOVERRIDE;
+    void OnExit() wxOVERRIDE {}
 
 private:
-    DECLARE_DYNAMIC_CLASS(wxGtkPrintModule)
+    wxDECLARE_DYNAMIC_CLASS(wxGtkPrintModule);
 };
 
 bool wxGtkPrintModule::OnInit()
 {
-#ifndef __WXGTK3__
-    if (gtk_check_version(2,10,0) == NULL)
-#endif
+    if (wx_is_at_least_gtk2(10))
     {
         wxPrintFactory::SetPrintFactory( new wxGtkPrintFactory );
     }
     return true;
 }
 
-IMPLEMENT_DYNAMIC_CLASS(wxGtkPrintModule, wxModule)
+wxIMPLEMENT_DYNAMIC_CLASS(wxGtkPrintModule, wxModule);
 
 //----------------------------------------------------------------------------
 // wxGtkPrintFactory
@@ -385,7 +377,7 @@ extern "C"
 // wxGtkPrintNativeData
 //----------------------------------------------------------------------------
 
-IMPLEMENT_CLASS(wxGtkPrintNativeData, wxPrintNativeDataBase)
+wxIMPLEMENT_CLASS(wxGtkPrintNativeData, wxPrintNativeDataBase);
 
 wxGtkPrintNativeData::wxGtkPrintNativeData()
 {
@@ -399,6 +391,20 @@ wxGtkPrintNativeData::~wxGtkPrintNativeData()
     g_object_unref(m_config);
 }
 
+void wxGtkPrintNativeData::SetPrintJob(GtkPrintOperation* job)
+{
+    m_job = job;
+#if GTK_CHECK_VERSION(2,18,0)
+    if (job)
+    {
+        if (wx_is_at_least_gtk2(18))
+        {
+            gtk_print_operation_set_embed_page_setup(job, true);
+        }
+    }
+#endif
+}
+
 // Convert datas stored in m_config to a wxPrintData.
 // Called by wxPrintData::ConvertFromNative().
 bool wxGtkPrintNativeData::TransferTo( wxPrintData &data )
@@ -568,7 +574,13 @@ bool wxGtkPrintNativeData::TransferFrom( const wxPrintData &data )
 void wxGtkPrintNativeData::SetPrintConfig( GtkPrintSettings * config )
 {
     if (config)
+    {
+        if ( m_config )
+        {
+            g_object_unref(m_config);
+        }
         m_config = gtk_print_settings_copy(config);
+    }
 }
 
 // Extract page setup from settings.
@@ -598,7 +610,7 @@ void wxGtkPrintNativeData::SetPageSetupToSettings(GtkPrintSettings* settings, Gt
 // wxGtkPrintDialog
 //----------------------------------------------------------------------------
 
-IMPLEMENT_CLASS(wxGtkPrintDialog, wxPrintDialogBase)
+wxIMPLEMENT_CLASS(wxGtkPrintDialog, wxPrintDialogBase);
 
 wxGtkPrintDialog::wxGtkPrintDialog( wxWindow *parent, wxPrintDialogData *data )
                     : wxPrintDialogBase(parent, wxID_ANY, _("Print"),
@@ -680,7 +692,9 @@ int wxGtkPrintDialog::ShowModal()
     // If the settings are OK, we restore it.
     if (settings != NULL)
         gtk_print_operation_set_print_settings (printOp, settings);
-    gtk_print_operation_set_default_page_setup (printOp, native->GetPageSetupFromSettings(settings));
+    GtkPageSetup* pgSetup = native->GetPageSetupFromSettings(settings);
+    gtk_print_operation_set_default_page_setup (printOp, pgSetup);
+    g_object_unref(pgSetup);
 
     // Show the dialog if needed.
     GError* gError = NULL;
@@ -755,11 +769,16 @@ int wxGtkPrintDialog::ShowModal()
     return wxID_OK;
 }
 
+wxDC* wxGtkPrintDialog::GetPrintDC()
+{
+    return new wxPrinterDC(m_printDialogData.GetPrintData());
+}
+
 //----------------------------------------------------------------------------
 // wxGtkPageSetupDialog
 //----------------------------------------------------------------------------
 
-IMPLEMENT_CLASS(wxGtkPageSetupDialog, wxPageSetupDialogBase)
+wxIMPLEMENT_CLASS(wxGtkPageSetupDialog, wxPageSetupDialogBase);
 
 wxGtkPageSetupDialog::wxGtkPageSetupDialog( wxWindow *parent,
                             wxPageSetupDialogData* data )
@@ -819,6 +838,8 @@ int wxGtkPageSetupDialog::ShowModal()
     gtk_page_setup_unix_dialog_set_page_setup(
         GTK_PAGE_SETUP_UNIX_DIALOG(dlg), oldPageSetup);
 
+    g_object_unref(oldPageSetup);
+
     int result = gtk_dialog_run(GTK_DIALOG(dlg));
     gtk_widget_hide(dlg);
 
@@ -881,7 +902,7 @@ int wxGtkPageSetupDialog::ShowModal()
 // wxGtkPrinter
 //----------------------------------------------------------------------------
 
-IMPLEMENT_CLASS(wxGtkPrinter, wxPrinterBase)
+wxIMPLEMENT_CLASS(wxGtkPrinter, wxPrinterBase);
 
 wxGtkPrinter::wxGtkPrinter( wxPrintDialogData *data ) :
     wxPrinterBase( data )
@@ -944,6 +965,7 @@ bool wxGtkPrinter::Print(wxWindow *parent, wxPrintout *printout, bool prompt )
     dataToSend.printer = this;
     dataToSend.printout = printout;
 
+    wxDELETE(m_dc);
     // These Gtk signals are caught here.
     g_signal_connect (printOp, "begin-print", G_CALLBACK (gtk_begin_print_callback), &dataToSend);
     g_signal_connect (printOp, "draw-page", G_CALLBACK (gtk_draw_page_print_callback), &dataToSend);
@@ -951,7 +973,6 @@ bool wxGtkPrinter::Print(wxWindow *parent, wxPrintout *printout, bool prompt )
 
     // This is used to setup the DC and
     // show the dialog if desired
-    dialog.SetPrintDC(m_dc);
     dialog.SetShowDialog(prompt);
 
     // doesn't necessarily show
@@ -965,6 +986,9 @@ bool wxGtkPrinter::Print(wxWindow *parent, wxPrintout *printout, bool prompt )
         sm_lastError = wxPRINTER_ERROR;
     }
 
+    printout->SetDC(NULL);
+    wxDELETE(m_dc);
+
     return (sm_lastError == wxPRINTER_NO_ERROR);
 }
 
@@ -995,19 +1019,8 @@ void wxGtkPrinter::BeginPrint(wxPrintout *printout, GtkPrintOperation *operation
         return;
     }
 
-    printout->SetPPIScreen(wxGetDisplayPPI());
-    printout->SetPPIPrinter( printDC->GetResolution(),
-                             printDC->GetResolution() );
+    printout->SetUp(*m_dc);
 
-    printout->SetDC(m_dc);
-
-    int w, h;
-    m_dc->GetSize(&w, &h);
-    printout->SetPageSizePixels((int)w, (int)h);
-    printout->SetPaperRectPixels(wxRect(0, 0, w, h));
-    int mw, mh;
-    m_dc->GetSizeMM(&mw, &mh);
-    printout->SetPageSizeMM((int)mw, (int)mh);
     printout->OnPreparePrinting();
 
     // Get some parameters from the printout, if defined.
@@ -1137,7 +1150,6 @@ wxDC* wxGtkPrinter::PrintDialog( wxWindow *parent )
 {
     wxGtkPrintDialog dialog( parent, &m_printDialogData );
 
-    dialog.SetPrintDC(m_dc);
     dialog.SetShowDialog(true);
 
     int ret = dialog.ShowModal();
@@ -1153,9 +1165,13 @@ wxDC* wxGtkPrinter::PrintDialog( wxWindow *parent )
         return NULL;
     }
 
-    m_printDialogData = dialog.GetPrintDialogData();
+    wxDC* dc = dialog.GetPrintDC();
+    if ( dc )
+    {
+        m_printDialogData = dialog.GetPrintDialogData();
+    }
 
-    return new wxPrinterDC( m_printDialogData.GetPrintData() );
+    return dc;
 }
 
 bool wxGtkPrinter::Setup( wxWindow * WXUNUSED(parent) )
@@ -1186,13 +1202,12 @@ bool wxGtkPrinter::Setup( wxWindow * WXUNUSED(parent) )
 
 #endif
 
-IMPLEMENT_ABSTRACT_CLASS(wxGtkPrinterDCImpl, wxDCImpl)
+wxIMPLEMENT_ABSTRACT_CLASS(wxGtkPrinterDCImpl, wxDCImpl);
 
 wxGtkPrinterDCImpl::wxGtkPrinterDCImpl(wxPrinterDC *owner, const wxPrintData& data)
                   : wxDCImpl( owner )
+    , m_printData(data)
 {
-    m_printData = data;
-
     wxGtkPrintNativeData *native =
         (wxGtkPrintNativeData*) m_printData.GetNativeData();
 
@@ -1228,6 +1243,9 @@ wxGtkPrinterDCImpl::wxGtkPrinterDCImpl(wxPrinterDC *owner, const wxPrintData& da
 
 wxGtkPrinterDCImpl::~wxGtkPrinterDCImpl()
 {
+    if ( m_fontdesc )
+         pango_font_description_free(m_fontdesc);
+
     g_object_unref(m_context);
     g_object_unref(m_layout);
 }
@@ -1267,8 +1285,7 @@ void wxGtkPrinterDCImpl::DoGradientFillConcentric(const wxRect& rect, const wxCo
     wxCoord w =  rect.width;
     wxCoord h = rect.height;
 
-    const double r2 = (w/2)*(w/2)+(h/2)*(h/2);
-    double radius = sqrt(r2);
+    const double radius = wxMin(w, h) / 2.0;
 
     unsigned char redI = initialColour.Red();
     unsigned char blueI = initialColour.Blue();
@@ -1431,11 +1448,17 @@ void wxGtkPrinterDCImpl::DoDrawArc(wxCoord x1,wxCoord y1,wxCoord x2,wxCoord y2,w
         cairo_close_path (m_cairo);
 
         SetBrush( m_brush );
-        cairo_fill_preserve( m_cairo );
+        if ( m_pen.IsTransparent() )
+            cairo_fill(m_cairo);
+        else
+           cairo_fill_preserve(m_cairo);
     }
 
-    SetPen (m_pen);
-    cairo_stroke( m_cairo );
+    SetPen(m_pen);
+    if ( m_pen.IsNonTransparent() )
+    {
+        cairo_stroke(m_cairo);
+    }
 
     CalcBoundingBox (x1, y1);
     CalcBoundingBox (xc, yc);
@@ -1530,10 +1553,16 @@ void wxGtkPrinterDCImpl::DoDrawPolygon(int n, const wxPoint points[],
     cairo_close_path(m_cairo);
 
     SetBrush( m_brush );
-    cairo_fill_preserve( m_cairo );
+    if ( m_pen.IsTransparent() )
+        cairo_fill(m_cairo);
+    else
+        cairo_fill_preserve(m_cairo);
 
-    SetPen (m_pen);
-    cairo_stroke( m_cairo );
+    SetPen(m_pen);
+    if ( m_pen.IsNonTransparent() )
+    {
+        cairo_stroke(m_cairo);
+    }
 
     CalcBoundingBox( x, y );
 
@@ -1549,17 +1578,27 @@ void wxGtkPrinterDCImpl::DoDrawPolyPolygon(int n, const int count[], const wxPoi
 
 void wxGtkPrinterDCImpl::DoDrawRectangle(wxCoord x, wxCoord y, wxCoord width, wxCoord height)
 {
-    width--;
-    height--;
+    if ( m_pen.IsNonTransparent() )
+    {
+        // outline is one pixel larger than what raster-based wxDC implementations draw
+        width -= 1;
+        height -= 1;
+    }
 
     cairo_new_path(m_cairo);
     cairo_rectangle ( m_cairo, XLOG2DEV(x), YLOG2DEV(y), XLOG2DEVREL(width), YLOG2DEVREL(height));
 
     SetBrush( m_brush );
-    cairo_fill_preserve( m_cairo );
+    if ( m_pen.IsTransparent() )
+        cairo_fill(m_cairo);
+    else
+        cairo_fill_preserve(m_cairo);
 
-    SetPen (m_pen);
-    cairo_stroke( m_cairo );
+    SetPen(m_pen);
+    if ( m_pen.IsNonTransparent() )
+    {
+        cairo_stroke(m_cairo);
+    }
 
     CalcBoundingBox( x, y );
     CalcBoundingBox( x + width, y + height );
@@ -1604,10 +1643,16 @@ void wxGtkPrinterDCImpl::DoDrawRoundedRectangle(wxCoord x, wxCoord y, wxCoord wi
     cairo_close_path(m_cairo);
 
     SetBrush(m_brush);
-    cairo_fill_preserve(m_cairo);
+    if ( m_pen.IsTransparent() )
+        cairo_fill(m_cairo);
+    else
+        cairo_fill_preserve(m_cairo);
 
     SetPen(m_pen);
-    cairo_stroke(m_cairo);
+    if ( m_pen.IsNonTransparent() )
+    {
+        cairo_stroke(m_cairo);
+    }
 
     CalcBoundingBox(x,y);
     CalcBoundingBox(x+width,y+height);
@@ -1627,10 +1672,16 @@ void wxGtkPrinterDCImpl::DoDrawEllipse(wxCoord x, wxCoord y, wxCoord width, wxCo
     cairo_arc ( m_cairo, 0, 0, XLOG2DEVREL(width/2), 0, 2 * M_PI);
 
     SetBrush( m_brush );
-    cairo_fill_preserve( m_cairo );
+    if ( m_pen.IsTransparent() )
+        cairo_fill(m_cairo);
+    else
+        cairo_fill_preserve(m_cairo);
 
-    SetPen (m_pen);
-    cairo_stroke( m_cairo );
+    SetPen(m_pen);
+    if ( m_pen.IsNonTransparent() )
+    {
+        cairo_stroke(m_cairo);
+    }
 
     CalcBoundingBox( x, y );
     CalcBoundingBox( x + width, y + height );
@@ -1643,7 +1694,7 @@ void wxGtkPrinterDCImpl::DoDrawSpline(const wxPointList *points)
 {
     SetPen (m_pen);
 
-    double c, d, x1, y1, x2, y2, x3, y3;
+    double c, d, x1, y1, x3, y3;
     wxPoint *p, *q;
 
     wxPointList::compatibility_iterator node = points->GetFirst();
@@ -1670,6 +1721,7 @@ void wxGtkPrinterDCImpl::DoDrawSpline(const wxPointList *points)
     node = node->GetNext();
     while (node)
     {
+        double x2, y2;
         q = node->GetData();
 
         x1 = x3;
@@ -1898,11 +1950,11 @@ void wxGtkPrinterDCImpl::SetPen( const wxPen& pen )
     double width;
 
     if (m_pen.GetWidth() <= 0)
-        width = 0.1;
+        width = 0.1; // Thin, scale-independent line.
     else
-        width = (double) m_pen.GetWidth();
+        width = (double) m_pen.GetWidth() * m_scaleX;
 
-    cairo_set_line_width( m_cairo, width * m_DEV2PS * m_scaleX );
+    cairo_set_line_width( m_cairo, width * m_DEV2PS );
     static const double dotted[] = {2.0, 5.0};
     static const double short_dashed[] = {4.0, 4.0};
     static const double long_dashed[] = {4.0, 8.0};
@@ -2109,6 +2161,8 @@ void wxGtkPrinterDCImpl::DoSetClippingRegion(wxCoord x, wxCoord y, wxCoord width
 void wxGtkPrinterDCImpl::DestroyClippingRegion()
 {
     cairo_reset_clip(m_cairo);
+
+    wxDCImpl::DestroyClippingRegion();
 }
 
 bool wxGtkPrinterDCImpl::StartDoc(const wxString& WXUNUSED(message))
@@ -2232,6 +2286,38 @@ void wxGtkPrinterDCImpl::DoGetTextExtent(const wxString& string, wxCoord *width,
     cairo_restore( m_cairo );
 }
 
+bool wxGtkPrinterDCImpl::DoGetPartialTextExtents(const wxString& text, wxArrayInt& widths) const
+{
+    widths.Empty();
+
+    const wxCharBuffer data = text.utf8_str();
+    int w = 0;
+    if ( data.length() > 0 )
+    {
+        cairo_save(m_cairo);
+        cairo_scale(m_cairo, m_scaleX, m_scaleY);
+
+        pango_layout_set_text(m_layout, data, data.length());
+        PangoLayoutIter* iter = pango_layout_get_iter(m_layout);
+        do
+        {
+            PangoRectangle rect;
+            pango_layout_iter_get_cluster_extents(iter, NULL, &rect);
+            w += rect.width;
+            widths.Add(PANGO_PIXELS(w));
+        } while (pango_layout_iter_next_cluster(iter));
+        pango_layout_iter_free(iter);
+
+       cairo_restore(m_cairo);
+    }
+    size_t i = widths.GetCount();
+    const size_t len = text.length();
+    while (i++ < len)
+        widths.Add(PANGO_PIXELS(w));
+
+    return true;
+}
+
 void wxGtkPrinterDCImpl::DoGetSize(int* width, int* height) const
 {
     GtkPageSetup *setup = gtk_print_context_get_page_setup( m_gpc );
@@ -2282,7 +2368,7 @@ int wxGtkPrinterDCImpl::GetResolution() const
 // Print preview
 // ----------------------------------------------------------------------------
 
-IMPLEMENT_CLASS(wxGtkPrintPreview, wxPrintPreviewBase)
+wxIMPLEMENT_CLASS(wxGtkPrintPreview, wxPrintPreviewBase);
 
 void wxGtkPrintPreview::Init(wxPrintout * WXUNUSED(printout),
                              wxPrintout * WXUNUSED(printoutForPrinting),
