@@ -20,7 +20,11 @@
 
 #include "wx/utils.h"
 
-#define USE_PUTENV (!defined(HAVE_SETENV) && defined(HAVE_PUTENV))
+#if !defined(HAVE_SETENV) && defined(HAVE_PUTENV)
+    #define USE_PUTENV 1
+#else
+    #define USE_PUTENV 0
+#endif
 
 #ifndef WX_PRECOMP
     #include "wx/string.h"
@@ -46,7 +50,7 @@
 
 #include "wx/private/selectdispatcher.h"
 #include "wx/private/fdiodispatcher.h"
-#include "wx/unix/execute.h"
+#include "wx/unix/private/execute.h"
 #include "wx/unix/pipe.h"
 #include "wx/unix/private.h"
 
@@ -147,20 +151,10 @@
 #if !defined(HAVE_USLEEP) && \
     ((defined(__SUN__) && !defined(__SunOs_5_6) && \
                          !defined(__SunOs_5_7) && !defined(__SUNPRO_CC)) || \
-     defined(__osf__) || defined(__EMX__))
+     defined(__osf__))
     extern "C"
     {
-        #ifdef __EMX__
-            /* I copied this from the XFree86 diffs. AV. */
-            #define INCL_DOSPROCESS
-            #include <os2.h>
-            inline void usleep(unsigned long delay)
-            {
-                DosSleep(delay ? (delay/1000l) : 1l);
-            }
-        #else // Unix
-            int usleep(unsigned int usec);
-        #endif // __EMX__/Unix
+        int usleep(unsigned int usec);
     };
 
     #define HAVE_USLEEP 1
@@ -198,9 +192,6 @@ void wxMicroSleep(unsigned long microseconds)
     #endif // Sun
 
     usleep(microseconds);
-#elif defined(HAVE_SLEEP)
-    // under BeOS sleep() takes seconds (what about other platforms, if any?)
-    sleep(microseconds * 1000000);
 #else // !sleep function
     #error "usleep() or nanosleep() function required for wxMicroSleep"
 #endif // sleep function
@@ -253,6 +244,10 @@ int wxKill(long pid, wxSignal sig, wxKillError *rc, int flags)
 // Shutdown or reboot the PC
 bool wxShutdown(int flags)
 {
+#if defined(__WXOSX__) && wxOSX_USE_IPHONE
+    wxUnusedVar(flags);
+    return false;
+#else
     flags &= ~wxSHUTDOWN_FORCE;
 
     wxChar level;
@@ -276,6 +271,7 @@ bool wxShutdown(int flags)
     }
 
     return system(wxString::Format("init %c", level).mb_str()) == 0;
+#endif
 }
 
 // ----------------------------------------------------------------------------
@@ -305,14 +301,14 @@ bool wxPipeInputStream::CanRead() const
     {
         case -1:
             wxLogSysError(_("Impossible to get child process input"));
-            // fall through
+            wxFALLTHROUGH;
 
         case 0:
             return false;
 
         default:
             wxFAIL_MSG(wxT("unexpected select() return value"));
-            // still fall through
+            wxFALLTHROUGH;
 
         case 1:
             // input available -- or maybe not, as select() returns 1 when a
@@ -346,7 +342,7 @@ size_t wxPipeOutputStream::OnSysWrite(const void *buffer, size_t size)
 #endif
            // do not treat it as an error
            m_file->ClearLastError();
-           // fall through
+           wxFALLTHROUGH;
 
        // no error
        case 0:
@@ -410,12 +406,12 @@ public:
 
         for ( int i = 0; i < m_argc; i++ )
         {
-            m_argv[i] = wxStrdup(args[i]);
+            m_argv[i] = wxStrdup(args[i].mb_str(wxConvWhateverWorks));
         }
     }
 
 #if wxUSE_UNICODE
-    ArgsArray(wchar_t **wargv)
+    ArgsArray(const wchar_t* const* wargv)
     {
         int argc = 0;
         while ( wargv[argc] )
@@ -440,7 +436,7 @@ public:
         delete [] m_argv;
     }
 
-    operator char**() const { return m_argv; }
+    operator const char* const*() const { return m_argv; }
 
 private:
     void Init(int argc)
@@ -462,8 +458,8 @@ private:
 // wxExecute implementations
 // ----------------------------------------------------------------------------
 
-#if defined(__DARWIN__)
-bool wxMacLaunch(char **argv);
+#if defined(__DARWIN__) && !defined(__WXOSX_IPHONE__)
+bool wxCocoaLaunch(const char* const* argv, pid_t &pid);
 #endif
 
 long wxExecute(const wxString& command, int flags, wxProcess *process,
@@ -477,7 +473,7 @@ long wxExecute(const wxString& command, int flags, wxProcess *process,
 
 #if wxUSE_UNICODE
 
-long wxExecute(wchar_t **wargv, int flags, wxProcess *process,
+long wxExecute(const wchar_t* const* wargv, int flags, wxProcess* process,
         const wxExecuteEnv *env)
 {
     ArgsArray argv(wargv);
@@ -536,25 +532,25 @@ int BlockUntilChildExit(wxExecuteData& execData)
         stdoutHandler.reset(new wxExecuteFDIOHandler
                                 (
                                     dispatcher,
-                                    execData.fdOut,
-                                    execData.bufOut
+                                    execData.m_fdOut,
+                                    execData.m_bufOut
                                 ));
         stderrHandler.reset(new wxExecuteFDIOHandler
                                 (
                                     dispatcher,
-                                    execData.fdErr,
-                                    execData.bufErr
+                                    execData.m_fdErr,
+                                    execData.m_bufErr
                                 ));
     }
 #endif // wxUSE_STREAMS
 
     // And dispatch until the PID is reset from wxExecuteData::OnExit().
-    while ( execData.pid )
+    while ( execData.m_pid )
     {
         dispatcher.Dispatch();
     }
 
-    return execData.exitcode;
+    return execData.m_exitcode;
 #else // !wxUSE_SELECT_DISPATCHER
     wxFAIL_MSG( wxS("Can't block until child exit without wxSelectDispatcher") );
 
@@ -565,7 +561,7 @@ int BlockUntilChildExit(wxExecuteData& execData)
 } // anonymous namespace
 
 // wxExecute: the real worker function
-long wxExecute(char **argv, int flags, wxProcess *process,
+long wxExecute(const char* const* argv, int flags, wxProcess* process,
         const wxExecuteEnv *env)
 {
     // for the sync execution, we return -1 to indicate failure, but for async
@@ -586,15 +582,15 @@ long wxExecute(char **argv, int flags, wxProcess *process,
     wxASSERT_MSG( wxThread::IsMain(),
                     wxT("wxExecute() can be called only from the main thread") );
 #endif // wxUSE_THREADS
-
-#if defined(__WXCOCOA__) || ( defined(__WXOSX_MAC__) && wxOSX_USE_COCOA_OR_CARBON )
-    // wxMacLaunch() only executes app bundles and only does it asynchronously.
+    pid_t pid;
+#if defined(__DARWIN__) && !defined(__WXOSX_IPHONE__)
+    pid = -1;
+    // wxCocoaLaunch() only executes app bundles and only does it asynchronously.
     // It returns false if the target is not an app bundle, thus falling
     // through to the regular code for non app bundles.
-    if ( !(flags & wxEXEC_SYNC) && wxMacLaunch(argv) )
+    if ( !(flags & wxEXEC_SYNC) && wxCocoaLaunch(argv, pid) )
     {
-        // we don't have any PID to return so just make up something non null
-        return -1;
+        return pid;
     }
 #endif // __DARWIN__
 
@@ -602,8 +598,8 @@ long wxExecute(char **argv, int flags, wxProcess *process,
     wxScopedPtr<wxExecuteData> execDataPtr(new wxExecuteData);
     wxExecuteData& execData = *execDataPtr;
 
-    execData.flags = flags;
-    execData.process = process;
+    execData.m_flags = flags;
+    execData.m_process = process;
 
     // create pipes for inter process communication
     wxPipe pipeIn,      // stdin
@@ -642,9 +638,9 @@ long wxExecute(char **argv, int flags, wxProcess *process,
     //     But on OpenVMS we do not have fork so we have to use vfork and
     //     cross our fingers that it works.
 #ifdef __VMS
-   pid_t pid = vfork();
+   pid = vfork();
 #else
-   pid_t pid = fork();
+   pid = fork();
 #endif
    if ( pid == -1 )     // error?
     {
@@ -659,7 +655,7 @@ long wxExecute(char **argv, int flags, wxProcess *process,
         //     always opened so don't do it any more, after all there doesn't
         //     seem to be any real problem with keeping them opened
 
-#if !defined(__VMS) && !defined(__EMX__)
+#if !defined(__VMS)
         if ( flags & wxEXEC_MAKE_GROUP_LEADER )
         {
             // Set process group to child process' pid.  Then killing -pid
@@ -697,22 +693,13 @@ long wxExecute(char **argv, int flags, wxProcess *process,
         // the descriptors do not need to be closed but for now this is better
         // than never closing them at all as wx code never used FD_CLOEXEC.
 
-#ifdef __DARWIN__
         // TODO: Iterating up to FD_SETSIZE is both inefficient (because it may
         //       be quite big) and incorrect (because in principle we could
         //       have more opened descriptions than this number). Unfortunately
         //       there is no good portable solution for closing all descriptors
         //       above a certain threshold but non-portable solutions exist for
-        //       most platforms, see [http://stackoverflow.com/questions/899038/
+        //       most platforms, see [https://stackoverflow.com/questions/899038/
         //          getting-the-highest-allocated-file-descriptor]
-        //
-        // Unfortunately, we cannot do this safely on OS X, because libdispatch
-        // may crash when we do this:
-        //     Exception Type:  EXC_BAD_INSTRUCTION (SIGILL)
-        //     Exception Codes: 0x0000000000000001, 0x0000000000000000
-        //
-        //     Application Specific Information:
-        //     BUG IN CLIENT OF LIBDISPATCH: Do not close random Unix descriptors
         for ( int fd = 0; fd < (int)FD_SETSIZE; ++fd )
         {
             if ( fd != STDIN_FILENO  &&
@@ -722,7 +709,6 @@ long wxExecute(char **argv, int flags, wxProcess *process,
                 close(fd);
             }
         }
-#endif // !__DARWIN__
 
 
         // Process additional options if we have any
@@ -759,10 +745,10 @@ long wxExecute(char **argv, int flags, wxProcess *process,
             }
         }
 
-        execvp(*argv, argv);
+        execvp(*argv, const_cast<char**>(argv));
 
         fprintf(stderr, "execvp(");
-        for ( char **a = argv; *a; a++ )
+        for (const char* const* a = argv; *a; a++)
             fprintf(stderr, "%s%s", a == argv ? "" : ", ", *a);
         fprintf(stderr, ") failed with error %d!\n", errno);
 
@@ -781,8 +767,6 @@ long wxExecute(char **argv, int flags, wxProcess *process,
     }
     else // we're in parent
     {
-        execData.OnStart(pid);
-
         // prepare for IO redirection
 
 #if HAS_PIPE_STREAMS
@@ -818,11 +802,11 @@ long wxExecute(char **argv, int flags, wxProcess *process,
 
             if ( flags & wxEXEC_SYNC )
             {
-                execData.bufOut.Init(outStream);
-                execData.bufErr.Init(errStream);
+                execData.m_bufOut.Init(outStream);
+                execData.m_bufErr.Init(errStream);
 
-                execData.fdOut = fdOut;
-                execData.fdErr = fdErr;
+                execData.m_fdOut = fdOut;
+                execData.m_fdErr = fdErr;
             }
         }
 #endif // HAS_PIPE_STREAMS
@@ -834,17 +818,24 @@ long wxExecute(char **argv, int flags, wxProcess *process,
             pipeErr.Close();
         }
 
-        // For the asynchronous case we don't have to do anything else, just
-        // let the process run.
         if ( !(flags & wxEXEC_SYNC) )
         {
             // Ensure that the housekeeping data is kept alive, it will be
             // destroyed only when the child terminates.
             execDataPtr.release();
-
-            return execData.pid;
         }
 
+        // Put the housekeeping data into the child process lookup table.
+        // Note that when running asynchronously, if the child has already
+        // finished this call will delete the execData and call any
+        // wxProcess's OnTerminate() handler immediately.
+        execData.OnStart(pid);
+
+        // For the asynchronous case we don't have to do anything else, just
+        // let the process run (if not already finished).
+        if ( !(flags & wxEXEC_SYNC) )
+            return pid;
+
 
         // If we don't need to dispatch any events, things are relatively
         // simple and we don't need to delegate to wxAppTraits.
@@ -1084,18 +1075,28 @@ bool wxGetUserName(wxChar *buf, int sz)
 
 bool wxIsPlatform64Bit()
 {
-    const wxString machine = wxGetCommandOutput(wxT("uname -m"));
+#if SIZEOF_VOID_P == 8
+    (void)wxGetCommandOutput;
+    return true;  // 64-bit programs run only on 64-bit platforms
+#else
+    const wxString machine = wxGetCpuArchitectureName();
 
     // the test for "64" is obviously not 100% reliable but seems to work fine
     // in practice
     return machine.Contains(wxT("64")) ||
                 machine.Contains(wxT("alpha"));
+#endif
+}
+
+wxString wxGetCpuArchitectureName()
+{
+    return wxGetCommandOutput(wxT("uname -m"));
 }
 
 #ifdef __LINUX__
 
 static bool
-wxGetValueFromLSBRelease(wxString arg, const wxString& lhs, wxString* rhs)
+wxGetValueFromLSBRelease(const wxString& arg, const wxString& lhs, wxString* rhs)
 {
     // lsb_release seems to just read a global file which is always in UTF-8
     // and hence its output is always in UTF-8 as well, regardless of the
@@ -1126,26 +1127,33 @@ wxLinuxDistributionInfo wxGetLinuxDistributionInfo()
 }
 #endif // __LINUX__
 
-// these functions are in src/osx/utilsexc_base.cpp for wxMac
+// these functions are in src/osx/utils_base.mm for wxOSX.
 #ifndef __DARWIN__
 
-wxOperatingSystemId wxGetOsVersion(int *verMaj, int *verMin)
+wxOperatingSystemId wxGetOsVersion(int *verMaj, int *verMin, int *verMicro)
 {
     // get OS version
-    int major, minor;
+    int major = -1, minor = -1, micro = -1;
     wxString release = wxGetCommandOutput(wxT("uname -r"));
-    if ( release.empty() ||
-         wxSscanf(release.c_str(), wxT("%d.%d"), &major, &minor) != 2 )
+    if ( !release.empty() )
     {
-        // failed to get version string or unrecognized format
-        major =
-        minor = -1;
+        if ( wxSscanf(release.c_str(), wxT("%d.%d.%d"), &major, &minor, &micro ) != 3 )
+        {
+            micro = 0;
+            if ( wxSscanf(release.c_str(), wxT("%d.%d"), &major, &minor ) != 2 )
+            {
+                // failed to get version string or unrecognized format
+                major = minor = micro = -1;
+            }
+        }
     }
 
     if ( verMaj )
         *verMaj = major;
     if ( verMin )
         *verMin = minor;
+    if ( verMicro )
+        *verMicro = micro;
 
     // try to understand which OS are we running
     wxString kernel = wxGetCommandOutput(wxT("uname -s"));
@@ -1163,6 +1171,16 @@ wxString wxGetOsDescription()
     return wxGetCommandOutput(wxT("uname -s -r -m"));
 }
 
+bool wxCheckOsVersion(int majorVsn, int minorVsn, int microVsn)
+{
+    int majorCur, minorCur, microCur;
+    wxGetOsVersion(&majorCur, &minorCur, &microCur);
+
+    return majorCur > majorVsn
+        || (majorCur == majorVsn && minorCur >= minorVsn)
+        || (majorCur == majorVsn && minorCur == minorVsn && microCur >= microVsn);
+}
+
 #endif // !__DARWIN__
 
 unsigned long wxGetProcessId()
@@ -1203,7 +1221,12 @@ wxMemorySize wxGetFreeMemory()
                     {
                         unsigned long cached;
                         if ( sscanf(buf, "Cached: %lu", &cached) == 1 )
-                            memFree += cached;
+                        {
+                            if ( cached > ULONG_MAX-memFree )
+                                memFree = ULONG_MAX;
+                            else
+                                memFree += cached;
+                        }
                     }
 
                     // values here are always expressed in kB and we want bytes
@@ -1244,7 +1267,7 @@ bool wxGetDiskSpace(const wxString& path, wxDiskspaceSize_t *pTotal, wxDiskspace
 #if defined(HAVE_STATFS) || defined(HAVE_STATVFS)
     // the case to "char *" is needed for AIX 4.3
     wxStatfs_t fs;
-    if ( wxStatfs((char *)(const char*)path.fn_str(), &fs) != 0 )
+    if ( wxStatfs(const_cast<char*>(static_cast<const char*>(path.fn_str())), &fs) != 0 )
     {
         wxLogSysError( wxT("Failed to get file system statistics") );
 
@@ -1301,10 +1324,10 @@ public:
         gs_envVars.clear();
     }
 
-    DECLARE_DYNAMIC_CLASS(wxSetEnvModule)
+    wxDECLARE_DYNAMIC_CLASS(wxSetEnvModule);
 };
 
-IMPLEMENT_DYNAMIC_CLASS(wxSetEnvModule, wxModule)
+wxIMPLEMENT_DYNAMIC_CLASS(wxSetEnvModule, wxModule);
 
 #endif // USE_PUTENV
 
@@ -1386,7 +1409,8 @@ bool wxUnsetEnv(const wxString& variable)
 
 #include <signal.h>
 
-extern "C" void wxFatalSignalHandler(wxTYPE_SA_HANDLER)
+extern "C" {
+static void wxFatalSignalHandler(wxTYPE_SA_HANDLER)
 {
     if ( wxTheApp )
     {
@@ -1396,6 +1420,7 @@ extern "C" void wxFatalSignalHandler(wxTYPE_SA_HANDLER)
 
     abort();
 }
+}
 
 bool wxHandleFatalExceptions(bool doit)
 {
@@ -1480,8 +1505,8 @@ wxAppTraits::RunLoopUntilChildExit(wxExecuteData& execData,
     // It is possible that wxExecuteData::OnExit() had already been called
     // and reset the PID to 0, in which case we don't need to do anything
     // at all.
-    if ( !execData.pid )
-        return execData.exitcode;
+    if ( !execData.m_pid )
+        return execData.m_exitcode;
 
 #if wxUSE_STREAMS
     // Monitor the child streams if necessary.
@@ -1492,24 +1517,24 @@ wxAppTraits::RunLoopUntilChildExit(wxExecuteData& execData,
     {
         stdoutHandler.reset(new wxExecuteEventLoopSourceHandler
                                 (
-                                    execData.fdOut, execData.bufOut
+                                    execData.m_fdOut, execData.m_bufOut
                                 ));
         stderrHandler.reset(new wxExecuteEventLoopSourceHandler
                                 (
-                                    execData.fdErr, execData.bufErr
+                                    execData.m_fdErr, execData.m_bufErr
                                 ));
     }
 #endif // wxUSE_STREAMS
 
     // Store the event loop in the data associated with the child
     // process so that it could exit the loop when the child exits.
-    execData.syncEventLoop = &loop;
+    execData.m_syncEventLoop = &loop;
 
     // And run it.
     loop.Run();
 
     // The exit code will have been set when the child termination was detected.
-    return execData.exitcode;
+    return execData.m_exitcode;
 }
 
 // ----------------------------------------------------------------------------
@@ -1608,7 +1633,7 @@ void wxExecuteData::OnSomeChildExited(int WXUNUSED(sig))
     }
 }
 
-void wxExecuteData::OnStart(int pid_)
+void wxExecuteData::OnStart(int pid)
 {
     wxCHECK_RET( wxTheApp,
                  wxS("Ensure wxTheApp is set before calling wxExecute()") );
@@ -1622,29 +1647,41 @@ void wxExecuteData::OnStart(int pid_)
 
 
     // Remember the child PID to be able to wait for it later.
-    pid = pid_;
+    m_pid = pid;
 
     // Also save it in wxProcess where it will be accessible to the user code.
-    if ( process )
-        process->SetPid(pid);
+    if ( m_process )
+        m_process->SetPid(m_pid);
 
-    // Finally, add this object itself to the list of child processes so that
+    // Add this object itself to the list of child processes so that
     // we can check for its termination the next time we get SIGCHLD.
-    ms_childProcesses[pid] = this;
+    ms_childProcesses[m_pid] = this;
+
+    // However, if the child exited before we finished setting up above,
+    // we may have already missed its SIGCHLD.  So we also do an explicit
+    // check here before returning.
+    int exitcode;
+    if ( CheckForChildExit(m_pid, &exitcode) )
+    {
+        // Handle its termination if it did.
+        // This call will implicitly remove it from ms_childProcesses
+        // and, if running asynchronously, it will delete itself.
+        OnExit(exitcode);
+    }
 }
 
-void wxExecuteData::OnExit(int exitcode_)
+void wxExecuteData::OnExit(int exitcode)
 {
     // Remove this process from the hash list of child processes that are
     // still open as soon as possible to ensure we don't process it again even
     // if another SIGCHLD happens.
-    if ( !ms_childProcesses.erase(pid) )
+    if ( !ms_childProcesses.erase(m_pid) )
     {
-        wxFAIL_MSG(wxString::Format(wxS("Data for PID %d not in the list?"), pid));
+        wxFAIL_MSG(wxString::Format(wxS("Data for PID %d not in the list?"), m_pid));
     }
 
 
-    exitcode = exitcode_;
+    m_exitcode = exitcode;
 
 #if wxUSE_STREAMS
     if ( IsRedirected() )
@@ -1652,16 +1689,16 @@ void wxExecuteData::OnExit(int exitcode_)
         // Read the remaining data in a blocking way: this is fine because the
         // child has already exited and hence all the data must be already
         // available in the streams buffers.
-        bufOut.ReadAll();
-        bufErr.ReadAll();
+        m_bufOut.ReadAll();
+        m_bufErr.ReadAll();
     }
 #endif // wxUSE_STREAMS
 
     // Notify user about termination if required
-    if ( !(flags & wxEXEC_SYNC) )
+    if ( !(m_flags & wxEXEC_SYNC) )
     {
-        if ( process )
-            process->OnTerminate(pid, exitcode);
+        if ( m_process )
+            m_process->OnTerminate(m_pid, m_exitcode);
 
         // in case of asynchronous execution we don't need this object any more
         // after the child terminates
@@ -1670,10 +1707,10 @@ void wxExecuteData::OnExit(int exitcode_)
     else // sync execution
     {
         // let wxExecute() know that the process has terminated
-        pid = 0;
+        m_pid = 0;
 
         // Stop the event loop for synchronous wxExecute() if we're running one.
-        if ( syncEventLoop )
-            syncEventLoop->ScheduleExit();
+        if ( m_syncEventLoop )
+            m_syncEventLoop->ScheduleExit();
     }
 }
